from ntdsdotsqlite.utils import get_ESE_column_names, create_database
from ntdsdotsqlite.containers import containers_generator
from ntdsdotsqlite.accounts import account_generator
from ntdsdotsqlite.domain import get_domain_objects
from ntdsdotsqlite.groups import group_generator
from ntdsdotsqlite.orga_units import ou_generator
from ntdsdotsqlite.decrypt import decrypt_sqlite
from ntdsdotsqlite.links import compute_links
from dissect.esedb import EseDB
import sqlite3
import json


def run(ese_path, outpath, system_path):
    create_database(outpath)
    sqlite_db = sqlite3.connect(outpath)
    ese_db = EseDB(open(ese_path, "rb"))
    cursor = sqlite_db.cursor()
    # Getting column names
    number_rows, column_names = get_ESE_column_names(ese_path)
    ese_db.column_names = column_names
    # Compute links
    print("Retrieving and storing links information ...")
    link_relations = compute_links(ese_db)
    # Get the domain
    print("Retrieving the domain object ...")
    domain = get_domain_objects(ese_db)
    # Insert the domain record
    stmt = (
        "INSERT INTO domains VALUES(:id, :name, :netbios_name, :functional_level, :GUID, :gplink,"
        ":SID, :machineAccountQuota, :maxPwdAge, :lockoutDuration, :minPwdLength, :pwdHistoryLength"
        ", :minPwdAge, :dn)"
    )
    cursor.execute(stmt, domain)
    sqlite_db.commit()
    # Insert container objects
    print("Retrieving containers objects ...")
    stmt = """
        INSERT INTO containers VALUES (
        :id, :name, :description, :cn, :parent, :dn, :is_deleted
        )
    """
    cursor.executemany(stmt, containers_generator(ese_db, sqlite_db))
    sqlite_db.commit()
    # Insert ou objects records
    stmt = """
        INSERT INTO organizational_units VALUES (
        :id, :name, :description, :parent, :dn, :isDeleted
        )
    """
    print("Retrieving organizational units objects ...")
    cursor.executemany(stmt, ou_generator(ese_db))
    sqlite_db.commit()
    # Insert group objects
    print("Retrieving groups objects ...")
    stmt = f"""
        INSERT INTO groups VALUES (
        :id, :name, :cn, :samaccountname, :SID, {domain['id']},
        :is_deleted, :description, ""
        )
    """
    cursor.executemany(stmt, group_generator(ese_db))
    sqlite_db.commit()
    # # Handle groups memberships with themselves
    for row in cursor.execute("SELECT id FROM groups"):
        memberOf = []
        new_cur = sqlite_db.cursor()
        for link in link_relations[row[0]]:
            res = new_cur.execute(f"SELECT id FROM groups WHERE id={link}")
            res = res.fetchone()
            if res is not None:
                memberOf.append(link)
        new_cur.execute(
            "UPDATE groups SET memberOf=? WHERE id=?", (json.dumps(memberOf), row[0])
        )
    sqlite_db.commit()
    # Insert all user account records
    print("Retrieving user accounts objects ...")
    accounts_iter = account_generator(
        ese_db, "bf967aa7-0de6-11d0-a285-00aa003049e2",
        sqlite_db, link_relations
    )
    stmt = f"""
        INSERT INTO user_accounts VALUES (
        :id, :nthash, :lmhash, :UAC, :description, :lastLogonTimestamp,
        :pwdLastSet, :adminCount, :displayName, :GUID, :SID, :SPN,
        {domain['id']}, :UPN, :login, :samaccountname, :commonname,
        :supplementalCredentials, :lmPwdHistory, :ntPwdHistory, :accountExpires,
        :uac_flags, :parent, :dn, :isDeleted, :primaryGroup,
        :memberOf, :links, :isDisabled
        )
    """
    cursor.executemany(stmt, accounts_iter)
    sqlite_db.commit()
    # Insert all machine account records
    print("Retrieving machine accounts objects ...")
    machines_iter = account_generator(
        ese_db, "bf967a86-0de6-11d0-a285-00aa003049e2",
        sqlite_db, link_relations
    )
    stmt = f"""
        INSERT INTO machine_accounts VALUES (
        :id, :nthash, :lmhash, :UAC, :description, :lastLogonTimestamp,
        :pwdLastSet, :adminCount, :displayName, :GUID, :SID, :SPN,
        {domain['id']}, :UPN, :login, :samaccountname, :commonname,
        :supplementalCredentials, :lmPwdHistory, :ntPwdHistory, :accountExpires,
        :uac_flags, :parent, :dn, :isDeleted, :primaryGroup, :links,
        :isDisabled
        )
    """
    cursor.executemany(stmt, machines_iter)
    sqlite_db.commit()
    if system_path:
        print("Decrypting stuff with SYSTEM hive ...")
        decrypt_sqlite(sqlite_db, ese_path, system_path)
    if sqlite_db:
        sqlite_db.close()



from ntdsdotsqlite.domain import get_domain_objects
from ntdsdotsqlite.utils import get_schema_object
from ntdsdotsqlite.utils import escape_dn_chars


def ou_generator(ese_db):
    dnt_cat, _ = get_schema_object(ese_db, "bf967aa5-0de6-11d0-a285-00aa003049e2")
    datatable = ese_db.table("datatable")
    ous = filter(
        lambda row: (cat := row.get(ese_db.column_names["objectCategory"])) and cat == dnt_cat,
        datatable.records()
    )
    ous = {
        ou.get("DNT_col"): ou for ou in ous
    }

    domain = get_domain_objects(ese_db)
    domain_id = domain["id"]
    dn_suffix = domain["dn"]
    for ou_id, ou in ous.items():
        ou_object = {
            "id": ou_id,
            "description": ou.get(ese_db.column_names["description"]),
            "name": ou.get(ese_db.column_names["name"]),
            "parent": ou.get("PDNT_col"),
            "isDeleted": ou.get(ese_db.column_names["isDeleted"]) == 1
        }
        # get the full path of the OU
        dn_prefix = "OU=" + escape_dn_chars(ou_object["name"])
        cur_object = ou
        while True:
            parent_dnt = cur_object.get("PDNT_col")
            # if $ROOT_OBJECT" is reached or the current object is not an OU anymore
            if parent_dnt == domain_id:
                break
            parent = ous[parent_dnt]
            cur_object = parent
            name = parent.get(ese_db.column_names["name"])
            dn_prefix += "," + "OU=" + escape_dn_chars(name)
        ou_object["dn"] = f"{dn_prefix},{dn_suffix}"
        yield ou_object



from ntdsdotsqlite.utils import get_schema_object
from ntdsdotsqlite.utils import escape_dn_chars


def containers_generator(ese_db, sqlite_db):
    dnt_cat, _ = get_schema_object(ese_db, "bf967a8b-0de6-11d0-a285-00aa003049e2")
    datatable = ese_db.table("datatable")
    containers = filter(
        lambda row: (cat := row.get(ese_db.column_names["objectCategory"])) and cat == dnt_cat,
        datatable.records()
    )
    containers = {
        row.get("DNT_col"): {
            "id": row.get("DNT_col"),
            "name": row.get(ese_db.column_names["name"]),
            "cn": row.get(ese_db.column_names["cn"]),
            "description": row.get(ese_db.column_names["description"]),
            "parent": row.get("PDNT_col"),
            "is_deleted": row.get(ese_db.column_names["isDeleted"]) == 1
        } for row in containers
    }
    cur = sqlite_db.cursor()
    res = cur.execute("SELECT id, dn FROM domains")
    domain_id, domain_DN = res.fetchone()
    for _, container in containers.items():
        cur_object = container
        parts = [f"CN={escape_dn_chars(cur_object['cn'])}"]
        while cur_object["parent"] != domain_id and cur_object["parent"] in containers.keys():
            cur_object = containers[cur_object["parent"]]
            parts.append(f"CN={escape_dn_chars(cur_object['cn'])}")
        container["dn"] = ",".join([*parts, domain_DN])

    yield from containers.values()



from ntdsdotsqlite.utils import get_schema_object, raw_to_sid


def group_generator(ese_db):
    group_dnt, _ = get_schema_object(ese_db, "bf967a9c-0de6-11d0-a285-00aa003049e2")
    datatable = ese_db.table("datatable")
    groups = filter(
        lambda row: (cat := row.get(ese_db.column_names["objectCategory"])) and cat == group_dnt,
        datatable.records()
    )
    for group in groups:
        yield {
            "id": group.get("DNT_col"),
            "name": group.get(ese_db.column_names["name"]),
            "cn": group.get(ese_db.column_names["cn"]),
            "samaccountname": group.get(ese_db.column_names["sAMAccountName"]),
            "SID": raw_to_sid(group.get(ese_db.column_names["objectSid"])),
            "is_deleted": group.get(ese_db.column_names["isDeleted"]) == 1,
            "description": group.get(ese_db.column_names["description"])
        }



from pathlib import Path
import argparse


from ntdsdotsqlite.ntdsdotsqlite import run


def main():
    parser = argparse.ArgumentParser(
        prog="NTDS.sqlite",
        description=(
            "This tool helps dumping NTDS.DIT file to an SQLite database"
        )
    )
    parser.add_argument("NTDS", help="The NTDS.DIT file", type=Path)
    parser.add_argument(
        "--system", required=False,
        help=(
            "The SYSTEM hive to decrypt hashes. If not provided, hashes will "
            "be encrypted inside the sqlite database."
        )
    )
    parser.add_argument(
        "-o", "--outfile", required=True,
        help="The sqlite database. Example : NTDS.sqlite"
    )
    args = parser.parse_args()
    run(args.NTDS, args.outfile, args.system)


if __name__ == "__main__":
    main()



from collections import defaultdict


def compute_links(ese_db):
    links = ese_db.table("link_table")
    relations = defaultdict(list)
    for link in links.records():
        flink = link.get("link_DNT")
        blink = link.get("backlink_DNT")
        relations[blink].append(flink)
    return relations



from ntdsdotsqlite.utils import get_schema_object, raw_to_guid, raw_to_sid
from ntdsdotsqlite.utils import escape_dn_chars


def get_domain_objects(ese_db):
    datatable = ese_db.table("datatable")
    # Locate the ID of the domainDNS schema object
    domainDNS_dnt, _ = get_schema_object(ese_db, "19195a5b-6da0-11d0-afd3-00c04fd930c9")
    if domainDNS_dnt is None:
        print("no domainDNS object ...")
        exit(1)
    domain_object = None
    for row in datatable.records():
        if (
            (object_category := row.get(ese_db.column_names["objectCategory"])) and
            object_category == domainDNS_dnt and
            row.get(ese_db.column_names["msDS-Behavior-Version"])
        ):
            domain_object = row
            # We only need the first object that represents the main domain
            break
    # go up in the tree to get the full domain name
    cur_object = domain_object
    parents = []
    while True:
        # If we got back to the root object (guid '00'*16, name='$ROOT_OBJECT$\u0000'), break
        parent_dnt = cur_object.get("PDNT_col")
        if parent_dnt == 2:
            break
        for row in datatable.records():
            if row.get("DNT_col") == parent_dnt:
                parents.append(row)
                cur_object = row
                break
    func_levels = {
        0: "2000", 2: "2003", 3: "2008", 4: "2008R2",
        5: "2012", 6: "2012R2", 7: "2016"
    }
    domain = {
        "id": domain_object.get("DNT_col"),
        "netbios_name": domain_object.get(ese_db.column_names["name"]),
        "name": ".".join([x.get(ese_db.column_names["name"]) for x in [domain_object, *parents]]),
        "parents": parents,
        "functional_level": (
            func_levels[domain_object.get(ese_db.column_names["msDS-Behavior-Version"])]
        ),
        "GUID": raw_to_guid(domain_object.get(ese_db.column_names["objectGUID"])),
        "gplink": domain_object.get(ese_db.column_names["gPLink"]),
        "SID": raw_to_sid(domain_object.get(ese_db.column_names["objectSid"])),
        "machineAccountQuota": domain_object.get(ese_db.column_names["ms-DS-MachineAccountQuota"]),
        # max password age in seconds
        "maxPwdAge": (
            domain_object.get(ese_db.column_names["maxPwdAge"]) * -1 / 10000000
        ),
        # lockout duration in seconds
        "lockoutDuration": (
            domain_object.get(ese_db.column_names["lockoutDuration"]) * -1 / 10000000
        ),
        "minPwdLength": domain_object.get(ese_db.column_names["minPwdLength"]),
        "pwdHistoryLength": domain_object.get(ese_db.column_names["pwdHistoryLength"]),
        # minimum password age in seconds
        "minPwdAge": (
            domain_object.get(ese_db.column_names["minPwdAge"]) * -1 / 10000000
        ),
        "dn": "DC=" + ",DC=".join(
            [
                escape_dn_chars(x.get(ese_db.column_names["name"]))
                for x in [domain_object, *parents]
            ]
        )
    }
    return domain
